async def __init__(hub):
    hub.log.LOGGER = {}
    hub.log.HANDLER = []
    hub.log.INT_LEVEL = hub.lib.logging.INFO
    hub.log.QUEUE = hub.lib.asyncio.Queue()

    # Set up aliases for each log function
    hub.log.trace = hub.log.init.trace
    hub.log.info = hub.log.init.info
    hub.log.debug = hub.log.init.debug
    hub.log.error = hub.log.init.error
    hub.log.warning = hub.log.init.warning
    hub.log.warn = hub.log.warning
    hub.log.critical = hub.log.init.critical
    hub.log.fatal = hub.log.critical


async def get_caller(hub) -> tuple[str, int]:
    """
    Inspect the stack and get the function that called get_caller
    """
    if not __debug__:
        return "hub", 0

    try:
        # Grab the third ref back, fall back to closer refs
        ref, stack_summary = (hub._call_stack.get()[-3:] or [None])[0]
        filename, lineno, *_ = stack_summary
        return ref, lineno
    except Exception:
        return "hub", 0


async def get_logger(hub, name: str):
    """
    Create a logger for the given ref with all the configured handlers
    """
    if name not in hub.log.LOGGER:
        if hub.log.HANDLER:
            logger = hub.lib.aiologger.Logger(name=name, level=hub.log.INT_LEVEL)
            for handler in hub.log.HANDLER:
                handler.level = hub.log.INT_LEVEL
                logger.add_handler(handler)
        else:
            logger = hub.lib.aiologger.Logger.with_default_handlers(
                name=name, level=hub.log.INT_LEVEL
            )
        logger.emit = hub.log.init.emit

        hub.log.LOGGER[name] = logger

    return hub.log.LOGGER[name]


async def setup(
    hub, log_plugin: str = "init", *, log_level: str, log_file: str, **kwargs
):
    """
    Initialize the logger with the named plugin
    """
    if not __debug__:
        return

    if hub.log.HANDLER:
        # We already set up the logger
        return

    # Set up trace logger
    hub.lib.aiologger.levels.LEVEL_TO_NAME[5] = "TRACE"
    levels = hub.lib.logging.getLevelNamesMapping()
    levels["TRACE"] = 5
    hub.lib.logging.addLevelName(5, "TRACE")

    # Convert log level to integer
    if str(log_level).isdigit():
        hub.log.INT_LEVEL = int(log_level)
    else:
        hub.log.INT_LEVEL = levels[log_level.upper()]

    if log_plugin != "init":
        # Create the log file
        if log_file:
            path = await hub.lib.aiopath.Path(log_file).expanduser()
            if not await path.parent.exists():
                await path.parent.mkdir(parents=True, exist_ok=True)
            await path.touch(exist_ok=True)
            log_file = str(path)
        await hub.log[log_plugin].setup(log_file=log_file, **kwargs)

    if not __debug__:
        return

    # Create a handler that puts the main python log messages through aiologger via an asynchronous Queue
    class _AsyncHandler(hub.lib.logging.Handler):
        def emit(self, record):
            if hub.lib.asyncio._get_running_loop() is None:
                return
            hub._auto(hub.log.init.emit(record))

    async_handler = _AsyncHandler()

    # Replace all existing synchronous loggers with the aiologger
    for logger in hub.lib.logging.root.manager.loggerDict.values():
        if isinstance(logger, hub.lib.logging.Logger):
            for handler in logger.handlers[:]:
                logger.removeHandler(handler)
                handler.close()
            logger.addHandler(async_handler)
            logger.setLevel(hub.log.INT_LEVEL)


async def emit(hub, record):
    if not __debug__:
        return
    logger_name = f"lib.{record.name}"
    logger = await hub.log.init.get_logger(logger_name)
    message = record.getMessage()
    # Pass the log record to the appropriate aiologger instance
    hub._auto(
        await logger._log(
            record.levelno,
            message,
            (),
            extra={"lineno": record.lineno, "filepath": record.pathname},
        )
    )


async def trace(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger._log(5, msg, args, **kwargs, extra={"lineno": lineno})


async def debug(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger.debug(msg, *args, **kwargs, extra={"lineno": lineno})


async def info(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger.info(msg, *args, **kwargs, extra={"lineno": lineno})


async def warning(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger.warning(msg, *args, **kwargs, extra={"lineno": lineno})


async def error(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger.error(msg, *args, **kwargs, extra={"lineno": lineno})


async def critical(hub, msg: str, *args, **kwargs):
    if not __debug__:
        return
    ref, lineno = await hub.log.init.get_caller()
    logger = await hub.log.init.get_logger(ref)
    await logger.critical(msg, *args, **kwargs, extra={"lineno": lineno})



async def __virtual__(hub):
    """ """
    try:
        hub.opts  # pylint: disable=undefined-variable
    except Exception:  # pylint: disable=broad-except
        return False
    return True


async def present(hub):
    return True



async def pre(hub, ctx):
    print("default-pre")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["default-pre"]


async def post(hub, ctx):
    print("default-post")
    ctx.ret.append("default-post")


async def call(hub, ctx):
    print("default-pre-call")
    result = ctx.extra + [
        "default-pre-call",
        await ctx.func(hub=hub),
        "default-post-call",
    ]
    print("default-post-call")
    return result


async def pre_test_fn(hub, ctx):
    print("default-pre-test-fn")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["default-pre-test-fn"]


async def post_test_fn(hub, ctx):
    print("default-post-test-fn")
    ctx.ret.append("default-post-test-fn")


async def call_test_fn(hub, ctx):
    print("default-pre-call-test-fn")
    result = ctx.extra + [
        "default-pre-call-test-fn",
        await ctx.func(hub=hub),
        "default-post-call-test-fn",
    ]
    print("default-post-call-test-fn")
    return result



import asyncio
import importlib.machinery
import os
import secrets
import signal
import sys
import traceback
from contextvars import ContextVar, copy_context
from collections.abc import Iterable
from types import ModuleType
from typing import Any

import cpop.contract
import cpop.data
import cpop.dirs
import cpop.exc
import cpop.loader
import cpop.ref
import cpop.scanner
import cpop.verify


def ex_path(path) :
    """
    Take a path that is sent to the Sub and expand it if it is a string or not
    """
    if isinstance(path, str):
        return path.split(",")
    elif isinstance(path, list):
        return path
    else:
        return []


SHUTDOWN_SIGNAL = object()


class PopMeta(cpop.data.MultidictCache):
    _subs = None

    def __init__(self, data: list[dict[str, any]], parent: object):
        attrs = {}
        data.append(attrs)
        super().__init__(data, parent = parent)
        object.__setattr__(self, "_attrs", attrs)

    def __setitem__(self, key: str, value):
        # Recurse the dotted reference and set the value on the last part of the reference
        if "." in key:
            parts = key.split(".")
            finder = self
            for p in parts[:-1]:
                if not p:
                    continue
                if p not in finder:
                    finder[p] = cpop.data.NamespaceDict()
                finder = finder[p]

            finder[parts[len(parts)-1]] = value
        else:
            self._attrs[key] = value

    def __setattr__(self, key: str, value):
        if key.startswith("_"):
            object.__setattr__(self, key, value)
        else:
            self[key] = value
            self._clear()


class Hub(PopMeta):
    """
    The redistributed pop central hub. All components of the system are
    rooted to the Hub.
    """
    _call_stack = ContextVar("_call_stack", default=[])

    def __init__(
            self,
            load_all_dynes: bool = True,
            load_all_subdirs: bool = True,
            recurse_subdirs: bool = True,
            pop_mods: list[str] = None,
            cli: str = None,
            logs: bool = True,
            load_config: bool = True,
            ):
        self.__init_kwargs = {
            k: v for k, v in locals().items() if k != "self" and not k.startswith("_")
        }
        subs = {}
        sub_alias = {}
        imports = {}
        PopMeta.__init__(self, [subs, sub_alias, imports], parent=self)
        self._subs = subs
        self._sub_alias = sub_alias
        self._imports = imports
        self._dynamic = None
        self._dscan = False
        # Set up the conf OPT structure so it is always available
        self._opt = None
        self._task_count = 0
        self._aio_alive = {}

    async def __aenter__(
        self,
    ):
        self._tasks = asyncio.Queue()
        loop = asyncio.get_running_loop()

        # Add the dyne for python imports to live in to the hub
        if "lib" not in self._subs:
            self._subs["lib"] = await AsyncSub(
                self,
                subname="lib",
                dyne_name="lib",
            )
            # Allow the sub to find python modules dynamically
            self._subs["lib"] += sys.modules

        pop_mods = self.__init_kwargs["pop_mods"]
        if pop_mods is None:
            pop_mods = ["pop.mods"]
        if "pop" not in self._subs:

            # Add the pop sub to the hub, this should always use pypath and
            # Should never be made dynamic. This is a core system sub and should
            # NOT be app-merged
            self._subs["pop"] = await AsyncSub(self, subname="pop", pypath=pop_mods)
            await self._subs["pop"]._load_all()

        # Add passthrough sig handlers to the running looop.
        # Their behavior can be modified with contracts
        sigint = self.pop.task.sigint
        loop.add_signal_handler(signal.SIGINT, lambda s=signal.SIGINT: loop.create_task(sigint(s)))
        sigterm = self.pop.task.sigterm
        loop.add_signal_handler(signal.SIGTERM, lambda s=signal.SIGTERM: loop.create_task(sigterm(s)))

        self._load_all_dynes = self.__init_kwargs["load_all_dynes"]
        self._load_all_subdirs = self.__init_kwargs["load_all_subdirs"]
        self._recurse_subdirs = self.__init_kwargs["recurse_subdirs"]
        await self._scan_dynamic()
        await self.pop.sub.add(dyne_name="log")
        await self._subs["log"]._load_all()

        # Set up config first
        if self.__init_kwargs["load_config"]:
            # Overwrite opt with config data
            self._opt = await self.pop.config.load(cli=self.__init_kwargs["cli"], **self._dynamic.config)

        # Set up a logger
        if self.__init_kwargs["logs"]:
            await self.pop.sub.load_subdirs(self._subs["log"])
            await self.log.init.setup(**self._opt.log.copy())

        # Load all other dynes
        if self._load_all_dynes:
            # Load dynes after config so they can use hub.OPT in their __init__s
            await self._load_all()

        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        if exc_type is not None and exc_value is not None:
            self._traceback(exc_type, exc_value, tb)
        await self._tasks.put(SHUTDOWN_SIGNAL)

    @property
    def OPT(self):
        return self._opt

    @property
    def _(self):
        """
        Return the parent of the last function on the call stack
        """
        stack = self._call_stack.get()
        if stack:
            last_ref = stack[len(stack) - 1][0]
            last_mod = last_ref.rsplit(".", maxsplit=1)[0]
            return cpop.ref.last(self, last_mod)

        # If nothing else worked return an instance of the hub
        return self

    async def _load_all(self):
        for base, imp in self._dynamic.imports__.items():
            self._subs["lib"]._imports[base] = imp

        # Load all dynamic subs onto the hub
        for dyne in self._dynamic.dyne:
            if dyne in self._subs:
                continue
            await self.pop.sub.add(dyne_name=dyne)
            if not self._load_all_subdirs:
                continue
            await self.pop.sub.load_subdirs(
                self._subs[dyne], recurse=self._recurse_subdirs
            )

    async def _holder(self):
        """
        Just a sleeping while loop to hold the loop open while it runs until
        complete
        """
        while True:
            complete = await self._tasks.get()
            if complete is  SHUTDOWN_SIGNAL:
                break
            await complete["task"]
            if complete["cb"]:
                await self._auto(complete["cb"])
            self._task_count -= 1

    def _auto(self, coro, cb_coro=None):
        """
        Start a task that will be automatically awaited

        coro: An unscheduled coroutine object to run
        cb_coro: An unscheduled coroutine to run when the main coroutine completes
        """
        current_context = copy_context()
        current_stack = self._call_stack.get().copy()

        # Create a new context for the task to carry its own call stack
        token = self._call_stack.set(current_stack)
        try:
            self._task_count += 1

            task = asyncio.create_task(coro, context=current_context)

            def _callback(task):
                original_context = copy_context()
                original_context.run(self._tasks.put_nowait, {"task": task, "cb": cb_coro})
                try:
                    # This will re-raise any exception caught by the task
                    task.result()
                except Exception as e:
                    # Print the traceback for the task
                    tb = sys.exc_info()[2]
                    self._traceback(type(e), e, tb)

            task.add_done_callback(_callback)
            return task
        finally:
            # Cleanup, reset the context when add_done_callback
            self._call_stack.reset(token)

    @classmethod
    def _call_stack_push(cls, ref: str, stack_summary):
        stack = cls._call_stack.get()
        stack.append((ref, stack_summary))

    @classmethod
    def _call_stack_pop(cls):
        stack = cls._call_stack.get()
        if stack:
            stack.pop()

    def _traceback(self, exctype, value, tb):
        call_stack = self._call_stack.get()
        if call_stack:
            print("\nHub Call Stack:", file=sys.stderr)
            # Iterate over all calls in the custom stack
            for ref, stack in reversed(call_stack):  # reverse to print from earliest to most recent
                print(f"\nCall from hub.{ref}:", file=sys.stderr)
                for frame in reversed(stack):  # reverse to print from earliest to most recent frame
                    print(f"  File \"{frame.filename}\", line {frame.lineno}, in {frame.name}", file=sys.stderr)
                    line = frame.line.strip() if frame.line else None
                    if line:
                        print(f"    {line}", file=sys.stderr)

        # Print the standard traceback
        print("\nStandard exception traceback (most recent call last):", file=sys.stderr)
        traceback.print_exception(exctype, value, tb, limit=1, file=sys.stderr)

    def __getstate__(self) -> dict:
        attrs = {}
        for k, v in self._attrs.items():
            if isinstance(v, cpop.loader.BASE_TYPES):
                attrs[k] = v
            elif isinstance(v, dict):
                if all(isinstance(v2, cpop.loader.BASE_TYPES) for v2 in v.values()):
                    attrs[k] = v
            elif isinstance(v, Iterable):
                if all(isinstance(v2, cpop.loader.BASE_TYPES) for v2 in v):
                    attrs[k] = v
        state = dict(
            init_kwargs=self.__init_kwargs,
            subs={name: sub.__getstate__() for name, sub in self._subs.items()},
            aliases=self._sub_alias,
            OPT=cpop.data.unfreeze(self._opt),
            imports={subname: mod.__name__ for subname, mod in self._imports.items()},
            attrs=attrs,
        )
        return state

    def __reduce__(self):
        return Hub, (), self.__getstate__()

    def __setstate__(self, state: dict):
        subs = state["subs"]
        opt = state["OPT"]
        aliases = state["aliases"]

        for subname, modname in state["imports"].items():
            try:
                self._imports[subname] = importlib.import_module(modname)
            except ModuleNotFoundError:
                ...
        self._attrs.update(state["attrs"])
        self._sub_alias.update(aliases)
        self._opt = cpop.data.freeze(opt)

        # Schedule the hub's async init function to run later
        self._auto(self.__aenter__())
        for name, data in subs.items():
            if name in self._subs:
                sub = self._subs[name]
            else:
                sub = Sub(hub=self, root=self, **data["init_kwargs"])
                self._subs[name] = sub

            sub.__setstate__(data)

    async def _scan_dynamic(self):
        """
        Refresh the dynamic roots data used for loading app merge module roots
        """
        self._clear()
        self._dynamic = await cpop.dirs.dynamic_dirs()
        self._dscan = True

    def __getattr__(self, item: str):
        if item.startswith("_"):
            return self.__getattribute__(item)
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"Hub has no attribute '{item}'")


class Sub(PopMeta):
    """
    The pop object that contains the loaded module data
    """

    def __init__(
        self,
        hub: Hub,
        *,
        subname: str,
        root: PopMeta = None,
        pypath: list[str] = None,
        static: list[str] = None,
        contracts_pypath: list[str] = None,
        contracts_static: list[str] = None,
        default_contracts: list[str] = None,
        virtual: bool = True,
        dyne_name: str = None,
        omit_start: tuple[str] = ("_",),
        omit_end: tuple[str] = (),
        omit_func: bool = False,
        omit_class: bool = False,
        omit_vars: bool = False,
        mod_basename: str = "",
        stop_on_failures: bool = False,
        init: bool = True,
        is_contract: bool = False,
        sub_virtual: bool = True,
        recursive_contracts_static: list[str] = None,
        default_recursive_contracts: list[str] = None,
        parent: object = None,
    ):
        """
        :param hub: The redistributed pop central hub
        :param subname: The name that the sub is going to take on the hub
            if nothing else is passed, it is used as the pypath
        :param pypath: One or many python paths which will be imported
        :param static: Directories that can be explicitly passed
        :param contracts_pypath: Load additional contract paths
        :param contracts_static: Load additional contract paths from a specific directory
        :param default_contracts: Specifies that a specific contract plugin will be applied as a default to all plugins
        :param virtual: Toggle whether or not to process __virtual__ functions
        :param dyne_name: The dynamic name to use to look up paths to find plugins -- linked to conf.py
        :param omit_start: Allows you to pass in a tuple of characters that would omit the loading of any object
            I.E. Any function starting with an underscore will not be loaded onto a plugin
            (You should probably never change this)
        :param omit_end:Allows you to pass in a tuple of characters that would omit the loading of an object
            (You should probably never change this)
        :param omit_func: bool: Don't load any functions
        :param omit_class: bool: Don't load any classes
        :param omit_vars: bool: Don't load any vars
        :param mod_basename: str: Manipulate the location in sys.modules that the plugin will be loaded to.
            Allow plugins to be loaded into a separate namespace.
        :param stop_on_failures: If any module fails to load for any reason, stacktrace and do not continue
            loading this sub
        :param init: bool: determine whether or not we process __init__ functions
        :param is_contract: Specify whether or not this sub is a contract
        :param sub_virtual: bool: Recursively ignore this sub and it's subs
        """
        self.__init_kwargs = {
            k: v
            for k, v in locals().items()
            if k not in ("self", "hub", "root") and not k.startswith("_")
        }
        subs = {}
        sub_alias = {}
        imports = {}
        loaded = {}

        _root = root or hub
        self._parent = parent or _root

        super().__init__([loaded, imports, subs, sub_alias], parent=self._parent)
        self._subs = subs
        self._sub_alias = sub_alias
        self._imports = imports
        self._loaded = loaded
        self._reload_mods = {}
        self._alias = []
        self._loaded_all = False
        self._load_errors = {}
        self._vmap = {}
        self._hub = hub
        self._root = _root
        self.__name__ = self._subname = subname
        self._pypath = ex_path(pypath)
        self._static = ex_path(static)
        self._contracts_pypath = ex_path(contracts_pypath)
        self._contracts_static = ex_path(contracts_static)
        self._recursive_contracts_static = ex_path(recursive_contracts_static)
        self._default_recursive_contracts = default_recursive_contracts or []
        self._default_contracts = default_contracts or ()
        self._dyne_name = dyne_name
        self._virtual = virtual
        self._omit_start = omit_start
        self._sub_virtual = sub_virtual
        self._omit_end = omit_end
        self._omit_func = omit_func
        self._omit_class = omit_class
        self._omit_vars = omit_vars
        self._mod_basename = mod_basename
        self._stop_on_failures = stop_on_failures
        self._is_contract = is_contract
        self._process_init = init

    async def __ainit__(self, *args, **kwargs):
        await self._prepare()

    async def _prepare(self):
        self._dirs = await cpop.dirs.dir_list(
            self._pypath,
            self._static,
        )

        if self._dyne_name:
            await self._load_dyne()
        self._contract_dirs = await cpop.dirs.dir_list(
            self._contracts_pypath,
            self._contracts_static,
        )
        self._contract_dirs.extend(await cpop.dirs.inline_dirs(self._dirs, "contracts"))
        self._recursive_contract_dirs = await cpop.dirs.dir_list(
            [],
            self._recursive_contracts_static,
        )
        self._recursive_contract_dirs.extend(
            await cpop.dirs.inline_dirs(self._dirs, "recursive_contracts")
        )

        self._contract_subs = []
        for i in range(len(self._contract_dirs)):
            sub = await AsyncSub(
                hub=self._hub,
                parent=self,
                subname=f"{self._subname}._contract_subs[{i}]",
                static=[self._contract_dirs[i]],
                is_contract=True,
            )
            await sub._load_all()
            self._contract_subs.append(sub)

        self._recursive_contract_subs = []
        if self._recursive_contract_dirs:
            self._recursive_contract_dirs += getattr(
                self._root, "_recursive_contract_dirs", []
            )
            # child Subs look at the parent's _recursive_contract_dirs, so we can't modify it
            filtered_recursive_contract_dirs = sorted(
                set(self._recursive_contract_dirs) - set(self._contract_dirs)
            )
            for i in range(len(filtered_recursive_contract_dirs)):
                sub = await AsyncSub(
                    hub=self._hub,
                    parent=self,
                    subname=f"{self._subname}._recursive_contract_subs[{i}]",
                    static=[filtered_recursive_contract_dirs[i]],
                    is_contract=True,
                )
                await sub._load_all()
                self._recursive_contract_subs.append(sub)
        else:
            # copy parent's recursive contracts
            self._recursive_contract_dirs = getattr(
                self._root, "_recursive_contract_dirs", []
            )
            self._recursive_contract_subs = getattr(
                self._root, "_recursive_contract_subs", []
            )

        self._name_root = self._load_name_root()
        self._scan = cpop.scanner.scan(self._dirs)

    async def _load_dyne(self):
        """
        Load up the dynamic dirs for this sub
        """
        if not self._hub._dscan:
            await self._hub._scan_dynamic()
        for path in self._hub._dynamic.dyne.get(self._dyne_name, {}).get("paths", []):
            if path not in self._dirs:
                self._dirs.append(path)

        self._clear()

    def _load_name_root(self):
        """
        Generate the root of the name to be used to apply to the loaded modules
        """
        if self._pypath:
            return self._pypath[0]
        elif self._dirs:
            return secrets.token_hex()

    def __getstate__(self):
        attrs = {}
        for k, v in self._attrs.items():
            if isinstance(v, dict):
                if all(isinstance(v2, cpop.loader.BASE_TYPES) for v2 in v.values()):
                    attrs[k] = v
            elif isinstance(v, Iterable):
                if all(isinstance(v2, cpop.loader.BASE_TYPES) for v2 in v):
                    attrs[k] = v
            elif isinstance(v, cpop.loader.BASE_TYPES):
                attrs[k] = v

        init_kwargs = self.__init_kwargs.copy()

        return dict(
            init_kwargs=init_kwargs,
            loaded={
                item: dict(
                    iface=iface, bname=bname, state=self._loaded[item].__getstate__()
                )
                for item, (iface, bname) in self._reload_mods.items()
            },
            subs={name: sub.__getstate__() for name, sub in self._subs.items()},
            aliases=self._sub_alias,
            imports={subname: mod.__name__ for subname, mod in self._imports.items()},
            attrs=attrs,
        )

    def __reduce__(self):
        return AsyncSub, (), self.__getstate__()

    def __setstate__(self, state: dict):
        """
        Only a hub can truly restore a sub because of their interdependency
        """
        subs = state["subs"]
        aliases = state["aliases"]
        imports = state["imports"]
        loaded = state["loaded"]

        self._sub_alias.update(aliases)
        # self._attrs.update(state["attrs"])

        for subname, modname in imports.items():
            try:
                self._imports[subname] = importlib.import_module(modname)
            except ModuleNotFoundError:
                ...

        for name, data in subs.items():
            sub = Sub(hub=self._hub, root=self, **data["init_kwargs"])
            sub.__setstate__(data)
            self._subs[name] = sub

        # Initialize the Sub, put these functions on the hub to be gathered
        self._hub._auto(self.__ainit__())
        self._hub._auto(self._load_all())

        # Make sure all unique items make it back onto the loaded mod
        for item, data in loaded.items():
            try:
                self._loaded[item].__setstate__(data["state"])
            except KeyError:
                ...

    def __getattr__(self, item: str):
        """
        If the item should be loaded, load it, else serve it
        """
        if item.startswith("_"):
            return self.__getattribute__(item)
        elif "." in item:
            return cpop.ref.last(self._hub, item)

        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"'{self._subname}' has no attribute '{item}'")

    async def _sub_init(self):
        """
        Run load init.py for the sub, running '__init__' function if present
        """
        await self._find_mod("init", match_only=True)

    def _process_load_error(
        self, mod: ModuleType, skip_full_stop: bool = False
    ) -> bool:
        if not isinstance(mod, cpop.loader.LoadError):
            # This is not a LoadError, return now!
            return False

        if mod.edict["verror"]:
            return False

        return True

    async def _find_mod(self, item: str, match_only: bool = False):
        """
        Find the module named item
        :param item: The module to search for (then load) from any scanned interface
        :param match_only: return the loaded module
        :return a loaded mod_dict
        """
        for iface in self._scan:
            for bname in self._scan[iface]:
                if os.path.basename(bname) == item:
                    await self._load_item(iface, bname)
            if item in self._loaded:
                self._reload_mods[item] = (iface, bname)
                return self._loaded[item]
        if not match_only:
            for iface in self._scan:
                for bname in self._scan[iface]:
                    if self._scan[iface][bname].get("loaded"):
                        continue
                    await self._load_item(iface, bname)
                    if item in self._loaded:
                        self._reload_mods[item] = (iface, bname)
                        return self._loaded[item]
        # Let's see if the module being lookup is in the load errors dictionary
        if item in self._load_errors:
            # Return the LoadError
            return self._load_errors[item]

    async def _load_item(self, iface: str, bname: str):
        """
        Load the named basename
        :param iface: A scanned directory type
        :param bname: The base name of the python path of a module
        """
        # The mname is the name to give the module in python's sys.modules
        # This name must be unique for every loaded module, so we use the full
        # module path sans the file extention
        mname = self._scan[iface][bname]["path"].replace(os.sep, ".")
        mname = mname[mname.index(".") + 1 : mname.rindex(".")].strip(".")
        mod = cpop.loader.load_mod(
            mname,
            iface,
            self._scan[iface][bname]["path"],
        )
        await self._prep_mod(mod, iface, bname)

    def _process_vret(self, vret: dict[str, Any]) -> bool:
        """
        :param vret: The return from a __virtual__ or __sub_virtual__ function
        :return: True if there was an error, else false
        """
        if "error" in vret:
            # Virtual Errors should not full stop pop
            self._process_load_error(vret["error"], skip_full_stop=True)
            # Store the LoadError under the __virtualname__ if defined
            self._load_errors[vret["vname"]] = vret["error"]
            return True
        else:
            return False

    async def _prep_mod(self, mod: ModuleType, iface: str, bname: str):
        """
        Prepare the module!
        :param mod: A python module containing data
        :param iface: A scanned directory type
        :param bname: The base name of the python path of a module
        """
        if not self._sub_virtual:
            return
        else:
            vret = await cpop.loader.load_sub_virtual(self._hub, mod, bname, self._ref)
            if self._process_vret(vret):
                self._sub_virtual = False
                return

        vret = await cpop.loader.load_virtual(self._hub, mod, bname, self._ref)

        if self._process_vret(vret):
            return

        contracts = []
        for contract_sub in self._contract_subs:
            contracts.extend(
                cpop.contract.load_contract(
                    contract_sub, self._default_contracts, mod, vret["name"]
                )
            )
        recursive_contracts = []
        for recursive_contract_sub in self._recursive_contract_subs:
            recursive_contracts.extend(
                cpop.contract.load_contract(
                    recursive_contract_sub,
                    self._default_recursive_contracts,
                    mod,
                    vret["name"],
                )
            )
        name = vret["name"]
        mod_dict = await cpop.loader.prep_loaded_mod(
            self, mod, name, contracts, recursive_contracts
        )
        if name != "init":
            cpop.verify.contract(contracts + recursive_contracts, mod_dict)
        self._loaded[name] = mod_dict
        self._vmap[mod.__file__] = name
        # Let's mark the module as loaded
        self._scan[iface][bname]["loaded"] = True
        if self._process_init:
            # Now that the module has been added to the sub, call mod_init
            await cpop.loader.mod_init(self, mod, name)

    async def _load_all(self):
        """
        Load all modules found during the scan.

        .. attention:: This completely disables the lazy loader behavior of pop
        """
        if self._loaded_all is True:
            return
        for iface in self._scan:
            for bname in self._scan[iface]:
                if self._scan[iface][bname].get("loaded"):
                    continue
                await self._load_item(iface, bname)
        self._loaded_all = True

    @property
    def _ref(self) -> str:
        """
        Return a full path on the hub to this sub
        """
        ref = self._subname
        root = self._parent
        while hasattr(root, "_subname"):
            ref = f"{root._subname}.{ref}"
            root = root._parent
        return ref


cdef class AsyncInitWrapper:
    cdef object cls

    def __init__(self, cls):
        self.cls = cls

    async def __call__(self, *args, **kwargs):
        cdef object instance = self.cls(*args, **kwargs)
        await instance.__ainit__(*args, **kwargs)
        return instance


AsyncSub = AsyncInitWrapper(Sub)



from aiopath import Path

current_dir = Path(globals()["__file__"]).parent


async def test_repeated_recursive_contracts(hub):
    # A scenario similar to this could occur when recursive contracts
    # were defined on a parent Sub, and a different set of recursive contracts
    # including some of the same contracts were defined on a child Sub.
    await hub.pop.sub.add(
        pypath=["tests.regression.contract_masking.sub"],
        recursive_contracts_static=[
            current_dir / "contract1",
            current_dir / "contract2",
            current_dir / "contract2",
        ],
    )
    val = await hub.sub.test.func()
    assert len(val) == 2
    assert val == ["contract2", "contract1"]



"""
used to test the sub init system
"""

# pylint: disable=undefined-variable


async def __init__(hub):
    """
    Add a value to the context
    """
    hub.context["init"] = True
    await hub._.inited()


async def inited(hub):
    return "init" in hub.context



# pylint: disable=expression-not-assigned
import cpop.exc
import pytest


async def test_sub_alias(hub):
    await hub.pop.sub.add(pypath=["tests.alias"])
    assert await hub.alias.init.ping() is True
    assert await hub.red.init.ping() is True
    assert await hub.blue.init.ping() is True
    assert await hub.green.init.ping() is True


async def test_basic(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.mods.test.ping()
    assert await hub.mods.test.ping() == {}
    assert await hub.mods.test.demo() is False
    assert await hub.mods.test.ping() == await hub.mods.foo.bar()


async def test_subdirs(hub):
    await hub.pop.sub.add(pypath=["tests.sdirs"])
    await hub.pop.sub.load_subdirs(hub.sdirs)
    assert await hub.sdirs.test.ping()
    assert await hub.sdirs.l11.test.ping()
    assert await hub.sdirs.l12.test.ping()
    assert await hub.sdirs.l13.test.ping()


async def test_subdirs_recurse(hub):
    await hub.pop.sub.add(pypath=["tests.sdirs"])
    await hub.pop.sub.load_subdirs(hub.sdirs, recurse=True)
    assert await hub.sdirs.test.ping()
    assert await hub.sdirs.l11.test.ping()
    assert await hub.sdirs.l11.l2.test.ping()
    assert await hub.sdirs.l12.l2.test.ping()
    assert await hub.sdirs.l13.l2.test.ping()


async def test_getattr(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.mods.test.ping()
    assert await hub["mods.test.ping"]() == {}
    assert await hub.mods.test["demo"]() is False
    assert await hub.mods.test.ping() == await hub["mods.foo.bar"]()


async def test_nest(hub):
    """
    Test the ability to nest the subs in a deeper namespace
    """
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.pop.sub.add(pypath=["tests.mods.nest"], sub=hub.mods)
    assert await hub.mods.nest.basic.ret_true()


async def test_func_attrs(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    assert "bar" in hub.mods.test.attr.__dict__
    assert hub.mods.test.attr.func.bar is True
    assert hub.mods.test.attr.func.func is not hub.mods.test.attr.func


async def test_module_level_direct_call(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    with pytest.raises(Exception):
        await hub.mods.test.module_level_non_aliased_ping_call()
    assert await hub.mods.test.module_level_non_aliased_ping_call_fw_hub() == {}


async def test_contract(hub):
    await hub.pop.sub.add(pypath=["tests.mods"], contracts_pypath=["tests.contracts"])
    with pytest.raises(Exception):
        await hub.mods.test.ping(4)


async def test_inline_contract(hub):
    await hub.pop.sub.add(pypath=["tests.cmods"])
    assert await hub.cmods.ctest.cping()
    assert hub.CPING


async def test_no_contract(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    with pytest.raises(TypeError):
        await hub.mods.test.ping(4)


async def test_contract_manipulate(hub):
    await hub.pop.sub.add(pypath=["tests.mods"], contracts_pypath=["tests.contracts"])
    assert "override" in await hub.mods.all.list()
    assert "post called" in await hub.mods.all.list()
    assert "post" in await hub.mods.all.dict()


async def test_private_function_cross_access(hub):
    hub.opts = "OPTS!"
    await hub.pop.sub.add(pypath=["tests.mods"])
    # Let's make sure that the private function is not accessible through the sub
    with pytest.raises(AttributeError):
        await hub.mods.priv._private() == "OPTS!"

    # Let's confirm that the private function has access to the cross
    # objects
    assert await hub.mods.priv.public() == "OPTS!"


async def test_private_function_cross_access_with_contracts(hub):
    hub.opts = "OPTS!"
    await hub.pop.sub.add(pypath=["tests.mods"], contracts_pypath=["tests.contracts"])
    # Let's make sure that the private function is not accessible through the sub
    with pytest.raises(AttributeError):
        await hub.mods.priv._private() == "OPTS!"

    # Let's confirm that the private function has access to the cross objects
    assert await hub.mods.priv.public() == "OPTS!"


async def test_cross_in_virtual(hub):
    hub.opts = "OPTS!"
    await hub.pop.sub.add(pypath=["tests.mods"], contracts_pypath=["tests.contracts"])
    assert await hub.mods.virt.present() is True


async def test_non_module_functions_are_not_loaded(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.mods._load_all()
    assert "scan" not in dir(hub.mods.test)
    await hub.mods.test.call_scan() is True

    """
    Verify that the assigned module name inside of the python sys.modules
    is unique
    """
    await hub.pop.sub.add(dyne_name="dyne1")
    mname = await hub.dyne3.init.mod_name()
    assert not mname.startswith(".")


async def test_dyne(hub):
    await hub.pop.sub.add(dyne_name="dyne1")
    assert hub.dyne1.INIT
    assert hub.dyne2.INIT
    assert hub.dyne3.INIT
    assert await hub.dyne1.test.dyne_ping()
    assert await hub.dyne1.nest.nest_dyne_ping()
    assert await hub.dyne2.test.dyne_ping()
    assert await hub.dyne2.nest.nest_dyne_ping()
    assert await hub.dyne3.test.dyne_ping()
    assert await hub.dyne3.nest.nest_dyne_ping()


async def test_sub_virtual(hub):
    await hub.pop.sub.add(dyne_name="dyne4")
    await hub.pop.sub.load_subdirs(hub.dyne4)
    assert "nest" in hub.dyne4._subs
    assert "nest" not in hub.dyne4._loaded


async def test_dyne_nest(hub):
    await hub.pop.sub.add(dyne_name="dn1")
    await hub.pop.sub.load_subdirs(hub.dn1, recurse=True)
    assert await hub.dn1.nest.dn1.ping()
    assert await hub.dn1.nest.dn2.ping()
    assert await hub.dn1.nest.dn3.ping()
    assert await hub.dn1.nest.next.test.ping()
    assert await hub.dn1.nest.next.last.test.ping()


async def test_existing_dyne(hub):
    await hub.pop.sub.add(dyne_name="dn1")
    await hub.pop.sub.load_subdirs(hub.dn1, recurse=True)
    assert "nest" in hub.dn1._subs
    assert "nest" not in hub.dn1._loaded
    # Add the same dyne again without recursing it and verify that the recursive elements are still there
    await hub.pop.sub.add(dyne_name="dn1")
    assert "nest" in hub.dn1._subs
    assert "nest" not in hub.dn1._loaded


async def test_dyne_extend(hub):
    await hub.pop.sub.add(dyne_name="dn1")
    await hub.pop.sub.load_subdirs(hub.dn1, recurse=True)
    assert await hub.dn1.nest.over.in_dn1()
    assert await hub.dn1.nest.over.in_dn2()
    assert await hub.dn1.nest.over.in_dn3()


async def test_dyne_overwrite(hub):
    await hub.pop.sub.add(dyne_name="dn1")
    await hub.pop.sub.load_subdirs(hub.dn1, recurse=True)
    # Assure that the first instance of a function gets overwritten
    assert await hub.dn1.nest.over.source() == "dn2"


async def test_contract_signatures(hub):
    hub.LOAD_PASS = True
    hub.LOAD_FAIL = False
    # These functions should load no problem
    await hub.pop.sub.add(pypath=["tests.mods.contract_sig"])


async def test_contract_signature_fail(hub):
    hub.LOAD_PASS = False
    hub.LOAD_FAIL = True
    # These functions should load with sig failures
    with pytest.raises(cpop.exc.ContractSigException) as e:
        await hub.pop.sub.add(pypath=["tests.mods.contract_sig"])
    errs = e.value.args[0].splitlines()

    assert "Signature Errors" in errs[0]

    filtered_errs = set({e for e in errs[1:-1] if "Enforcing signature" not in e})
    assert filtered_errs == {
        "args: **kwargs are not permitted as a parameter",
        "kwargs: *args are not permitted as a parameter",
    }


async def test_reload(hub):
    await hub.pop.sub.add(dyne_name="dn1")
    assert await hub.pop.sub.reload("dn1")


async def test_reload_fail(hub):
    assert not await hub.pop.sub.reload("nonexistant")



import sys
from unittest import mock

import aiopath
import cpop
import pytest


@pytest.fixture()
async def hub():
    async with cpop.Hub(
        # Let each test manually add their structure
        load_all_dynes=False,
        load_all_subdirs=False,
        recurse_subdirs=False,
        logs=False,
        load_config=False,
    ) as hub:
        yield hub


@pytest.fixture(autouse=True, scope="session")
async def tpath():
    code_dir = await aiopath.Path(__file__).parent.parent.absolute()
    assert await code_dir.exists()

    tpath_dir = code_dir / "tests" / "tpath"
    assert await tpath_dir.exists()

    NEW_PATH = [str(code_dir), str(tpath_dir)]

    for p in sys.path:
        if p not in NEW_PATH:
            NEW_PATH.append(p)

    with mock.patch("sys.path", NEW_PATH):
        yield



import pytest


@pytest.fixture(scope="function")
async def hub(hub):
    await hub.pop.sub.add(dyne_name="cn")
    yield hub


async def test_pre_fail_sync(hub):
    with pytest.raises(
        Exception, match="Pre contract 'pre_func1' failed for function 'func1'"
    ):
        await hub.cn.test.func1()


async def test_pre_fail_sync_with_message(hub):
    with pytest.raises(Exception, match="custom message"):
        await hub.cn.test.func2()


async def test_pre_fail_async(hub):
    with pytest.raises(
        Exception, match="Pre contract 'pre_afunc1' failed for function 'afunc1'"
    ):
        await hub.cn.test.afunc1()


async def test_pre_fail_async_with_message(hub):
    with pytest.raises(Exception, match="custom message"):
        await hub.cn.test.afunc2()


async def test_pre_fail_agen(hub):
    with pytest.raises(
        Exception, match="Pre contract 'pre_agen1' failed for function 'agen1'"
    ):
        async for _ in hub.cn.test.agen1():
            ...


async def test_pre_fail_agen_with_message(hub):
    with pytest.raises(Exception, match="custom message"):
        async for _ in hub.cn.test.agen2():
            ...



__virtualname__ = "all"


async def pre(hub, ctx):
    if len(ctx.args) > 1:
        raise ValueError("No can haz args!")
    if ctx.kwargs:
        raise ValueError("No can haz kwargs!")


async def call(hub, ctx):
    return await ctx.func(*ctx.args, **ctx.kwargs)


async def call_list(hub, ctx):
    return ["override"]


async def post(hub, ctx):
    ret = ctx.ret
    if isinstance(ret, list):
        ret.append("post called")
    elif isinstance(ret, dict):
        ret["post"] = "called"
    return ret



"""
This plugin is used to track pop enabled background task capabilities
"""

from collections.abc import Callable, Coroutine


async def auto(hub, coro: Coroutine, cb_coro=None):
    """
    Start a task that will be automatically awaited

    coro: An unscheduled coroutine object to run
    cb_coro: An unscheduled coroutine to run when the main coroutine completes
    """
    task = hub._auto(coro, cb_coro)
    await hub.lib.asyncio.sleep(0)
    return task


async def wait(hub):
    """
    When running without a long-running loop, (hub.aio.loop.run instead of hub.aio.loop.start)
    you need to call this function to clean up any auto tasks that have been created in the
    background.
    This function can be called from anywhere, it will just wait for all auto tasks to
    complete
    """
    while not hub._tasks.empty():
        complete = await hub._tasks.get()
        await complete["task"]
        if complete["cb"]:
            await hub.pop.task.auto(complete["cb"])


async def keep_alive(hub, precision: int, func: Callable, **kwargs):
    """
    Make sure that the named function stays alive, if the resulting coroutine dies, restart the function.
    The operation is run in a separate auto task

    :param precision: The number in seconds to sleep between checks
    :param func: The async function reference to restart if it dies
    :param **kwargs: The keyword arguments to send to the function to start and keep alive
    """

    async def _keep_alive():
        f_id = hub.lib.uuid.uuid4().hex
        if f_id not in hub._aio_alive:
            hub._aio_alive[f_id] = {"task": None}
        while True:
            start = False
            if hub._aio_alive[f_id]["task"] is None:
                start = True
            elif hub._aio_alive[f_id]["task"].done():
                await hub._aio_alive[f_id]["task"]
                start = True
            if start:
                hub._aio_alive[f_id]["task"] = await hub.pop.task.auto(func(**kwargs))
            await hub.lib.asyncio.sleep(precision)

    await hub.pop.task.auto(_keep_alive())


async def sigint(hub, signal: int):
    """
    Passthrough sigint that can be overridden with a call contract
    """
    raise KeyboardInterrupt


async def sigterm(hub, signal: int):
    """
    Passthrough sigterm that can be overridden with a call contract
    """
    hub.lib.sys.exit(0)



"""
module used while testing mock hubs provided in 'testing'.
"""

__contracts__ = ["testing"]


async def noparam(hub):
    pass


async def echo(hub, param):
    return param


async def signature_func(hub, param1, param2="default"):
    pass


async def attr_func(hub):
    pass


attr_func.test = True
attr_func.__test__ = True


async def async_echo(hub, param):
    return param



# pylint: disable=expression-not-assigned
import pytest

# Import pop


@pytest.mark.asyncio
async def test_async_simple_contracts(hub):
    await hub.pop.sub.add(pypath=["tests.coro"], subname="coro")
    ret = await hub.coro.test.simple()
    assert ret is True
    assert hub.PRE
    assert hub.POST



async def __virtual__(hub):
    return True


async def __init__(hub):
    hub._.FOO = "bar"
    assert hub._.__name__.endswith(".dunder"), f"name is {hub._.__name__}, not dunder"


async def func(hub):
    # Make sure that things defined in the init show up properly
    assert hub._.FOO == "bar"
    assert hub._.FOO == hub.mods.dunder.FOO



CONFIG_KEYWORDS = (
    "os",
    "default",
)


async def load(
    hub,
    cli: str = None,
    cli_config: dict[str, object] = None,
    config: dict[str, object] = None,
    subcommands: dict[str, object] = None,
    global_clis: list[str] = None,
    parser_args: tuple = None,
):
    """
    Use the pop-config system to load up a fresh configuration for this project
    from the included conf.py file.
    """
    if cli_config is None:
        cli_config = hub._dynamic.config.get("cli_config") or {}

    # Get the plain config data that will tell us about OS vars and defaults
    if config is None:
        config = hub._dynamic.config.get("config") or {}

    # Merge config and cli_config
    full_config = hub.lib.cpop.data.update(cli_config, config, merge_lists=True)

    # These CLI namespaces will be added on top of any cli
    if global_clis is None:
        global_clis = ["log", "pop"]

    # Initialize the active cli, this is what will go into argparse
    active_cli = {}

    # Logging options and config file are part of the global namespace
    for gn in global_clis:
        active_cli.update(full_config.get(gn, {}).copy())

    if subcommands is None:
        subcommands = hub._dynamic.config.get("subcommands") or {}
    else:
        active_subcommands = subcommands

    active_subcommands = subcommands.get(cli, {})

    # Grab the named cli last so that it can override globals
    active_cli.update(full_config.get(cli, {}))

    # Add all the cli options to argparse and call hte parser
    cli_opts = await hub.pop.config.parse_cli(
        cli=cli,
        active_cli=active_cli,
        subcommands=active_subcommands,
        parser_args=parser_args,
    )

    # Load the config file parameter in the proper order
    pop_config = full_config.get("pop", {}).get("config") or {}
    config_file = (
        cli_opts.get("config")
        or hub.lib.os.environ.get("POP_CONFIG", pop_config.get("os"))
        # TODO let the "default" be a reference to a function that returns the default or add a key for default_ref
        or pop_config.get("default")
    )

    config_data = {}
    if config_file:
        config_file = hub.lib.aiopath.Path(config_file)
        if await config_file.exists():
            async with config_file.open("r") as fh:
                config_data = hub.lib.yaml.safe_load(await fh.read())

    opt = await hub.pop.config.prioritize(
        cli=cli,
        cli_opts=cli_opts,
        config=full_config,
        config_file_data=config_data,
        global_clis=global_clis,
    )

    return hub.lib.cpop.data.freeze(opt)


async def parse_cli(
    hub,
    cli: str,
    active_cli: dict[str, object],
    subcommands: dict[str, object],
    parser_args: tuple = None,
) -> dict[str, object]:
    """
    Create a parser and parse all the CLI options.

    Args:
        hub: The POP hub instance.
        cli (str): The name of the CLI being parsed.
        active_cli (dict): The active CLI configuration.
        subcommands (dict): The subcommands configuration.

    Returns:
        argparse.Namespace: The parsed CLI options.
    """
    if not cli:
        return {}

    # Create the main parser for the CLI
    parser = hub.lib.argparse.ArgumentParser(
        description=f"{cli.title().replace('_', ' ')} CLI Parser"
    )
    sparser = None
    subparsers = {}

    # Add subcommands to the parser
    for subcommand, opts in subcommands.items():
        if sparser is None:
            sparser = parser.add_subparsers(dest="SUBPARSER")
        if "description" not in opts:
            opts["description"] = (
                f"{cli.title().replace('_', ' ')} {subcommand.title().replace('_', ' ')} CLI Parser"
            )

        subparsers[subcommand] = sparser.add_parser(
            subcommand,
            **opts,
        )

    # Add CLI options to the parser
    groups = {}
    subparser_groups = {subcommand: {} for subcommand in subparsers}

    for name, namespace_opts in active_cli.items():
        opts = namespace_opts.copy()
        positional = opts.pop("positional", False)
        # Config keywords will be handled by "prioritize"
        for opt in CONFIG_KEYWORDS:
            opts.pop(opt, None)
        cli_name = name if positional else f"--{name.lower().replace('_', '-')}"

        choices = opts.pop("choices", ())
        if isinstance(choices, str):
            finder = hub
            for part in choices.split("."):
                finder = getattr(finder, part)

            opts["choices"] = list(finder)

        type_ = opts.pop("type", None)
        if type_:
            opts["type"] = eval(type_)

        options = opts.pop("options", ())
        arg_subparsers = opts.pop("subcommands", [])

        # Handle argument groups for top-level parser
        group_name = opts.pop("group", None)
        target_group = (
            groups.setdefault(group_name, parser.add_argument_group(group_name))
            if group_name
            else parser
        )
        if "__global__" in arg_subparsers or not arg_subparsers:
            target_group.add_argument(cli_name, *options, **opts)

        # Handle argument groups for subparsers
        for subcommand in arg_subparsers:
            if subcommand == "__global__":
                for subcmd, sparser in subparsers.items():
                    subparser_group = (
                        subparser_groups[subcmd].setdefault(
                            group_name, sparser.add_argument_group(group_name)
                        )
                        if group_name
                        else sparser
                    )
                    subparser_group.add_argument(cli_name, *options, **opts)
            elif subcommand in subparsers:
                subcmd = subcommand
                sparser = subparsers[subcommand]
                subparser_group = (
                    subparser_groups[subcmd].setdefault(
                        group_name, sparser.add_argument_group(group_name)
                    )
                    if group_name
                    else sparser
                )
                subparser_group.add_argument(cli_name, *options, **opts)

    return hub.lib.cpop.data.NamespaceDict(parser.parse_args(args=parser_args).__dict__)


PLACEHOLDER = object()


async def prioritize(
    hub,
    cli: str,
    cli_opts: dict[str, any],
    config: dict[str, any],
    config_file_data: dict[str, any],
    global_clis: list[str],
):
    """
    Prioritize configuration data from various sources.

    The order of priority is:
    1. CLI options (highest priority)
    2. Configuration file data
    3. OS environment variables
    4. Default values (lowest priority)
    5. Rewrite the root_dir option so running apps automatically changes dirs to user preferences

    Args:
        hub: The cpop.hub instance.
        cli (str): The name of the CLI being prioritized.
        cli_opts (dict): The parsed CLI options.
        config (dict): The configuration dictionary.
        config_file_data (dict): The data from the configuration file.

    Returns:
       cpop.data.ImmutableNamespaceDict: The prioritized configuration options.
    """
    opt = hub.lib.collections.defaultdict(dict)
    root_dir = None
    for namespace, args in config.items():
        for arg, data in args.items():
            # Initialize value to None
            value = None

            # 1. Check CLI options first
            if (namespace == cli or namespace in global_clis) and arg in cli_opts:
                value = cli_opts.get(arg)

            # 2. Check config file data if CLI option is not set
            if value is None:
                value = config_file_data.get(namespace, {}).get(arg, PLACEHOLDER)

            # Skip malformed config
            if not isinstance(data, dict):
                msg = f"Invalid data from config.yaml: {data}"
                raise TypeError(msg)

            # 3. Check OS environment variables if config file data is not set
            if value is PLACEHOLDER and "os" in data:
                value = hub.lib.os.environ.get(data["os"], PLACEHOLDER)

            # 4. Use default value if none of the above are set
            if value is PLACEHOLDER:
                value = data.get("default")
            if arg == "root_dir":
                root_dir = value
            # Set the value in the OPT dictionary
            opt[namespace][arg] = value
    await hub.pop.config.manage_paths(cli, opt, root_dir)

    opt["pop"]["subparser"] = cli_opts.get("SUBPARSER", "")
    opt["pop"]["global_clis"] = global_clis

    return hub.lib.cpop.data.freeze(opt)


async def manage_paths(hub, cli: str, opt: dict, root_dir: str):
    """
    If a root_dir has been specified, then we support rewriting roots,
    find all of the options for _dir, _path, and _file and update them
    """
    reroot = True if root_dir else False
    if not root_dir:
        root_dir = hub.lib.aiopath.Path(hub.lib.os.sep)
    if root_dir == hub.lib.os.sep:
        if hasattr(hub.lib.os, "geteuid") and hub.lib.os.geteuid() != 0 and reroot:
            # Using aiopath.Path on all paths makes testing on multiple OSes much easier.
            # Since we can just patch aiopath.Path to be aiopath.PureWindowsPath or aiopath.PurePosixPath
            root_dir = (await hub.lib.aiopath.Path("~").expanduser()) / f".{cli}"
        else:
            root_dir = hub.lib.aiopath.Path(root_dir)
    else:
        return

    for namespace, args in opt.items():
        for key, val in args.items():
            if key == "root_dir":
                opt[namespace][key] = str(root_dir)
                continue
            try:
                path = hub.lib.aiopath.Path(val)
                if (
                    key.endswith(("_dir", "_path", "_file"))
                    and await path.is_absolute()
                ):
                    # only update absolute paths for keys
                    #  ending in _dir, _path or _file
                    if await path.is_absolute():
                        # remove the root from the path, then join it to the main root
                        opt[namespace][key] = root_dir.joinpath(*path.parts[1:])
            except TypeError:
                continue



# pylint: disable=expression-not-assigned
# Import pytest
import cpop.exc
import pytest

# Import pop


async def _test_calling_load_error_raises_pop_error(hub):
    """
    In this test, pop will continue loading although, when trying to
    access a functions which should be accessible on the module, a
    PopError is raised.
    """
    await hub.pop.sub.add(pypath=["tests.mods"], stop_on_failures=True)
    with pytest.raises(cpop.exc.PopError, match="Failed to load python module"):
        await hub.mods.bad_import.func()


async def test_verror_does_not_overload_loaded_mod(hub):
    """
    This tests will load 2 mods under the vname virtualname, however, one of them
    will explicitly not load. This makes sure load errors to not shadow good mod loads
    """
    await hub.pop.sub.add(
        pypath=["tests.mods.same_vname"],
        subname="mods",
    )
    assert await hub.mods.vname.func() == "wha? Yep!"


async def test_module_doesnt_exist(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    with pytest.raises(AttributeError, match="'mods' has no attribute 'doesntexist'"):
        await hub.mods.doesntexist



async def __virtual__(hub):
    return getattr(hub, "LOAD_FAIL", False), "Plugin not enabled"


async def args(hub, **__): ...


async def kwargs(hub, *_): ...


async def args_kwargs(hub): ...


async def async_func(hub): ...



async def test_load_with_config_file(hub, tmp_path):
    # Test loading with a configuration file
    cli = "test_cli"
    cli_config = hub._dynamic.config.cli_config
    cli_config.update({"test_cli": {"option": {}}})
    config = {"test_cli": {"option": {"default": "default"}}}
    config_file = tmp_path / "config.yaml"
    config_file.write_text("test_cli:\n  option: file_value\n")
    OPT = await hub.pop.config.load(
        cli=cli,
        cli_config=cli_config,
        config=config,
        subcommands={},
        global_clis=["pop"],
        parser_args=[f"--config={config_file}"],
    )
    assert OPT[cli]["option"] == "file_value"


async def test_load_with_subcommands(hub):
    # Test loading with subcommands
    cli = "test_cli"
    cli_config = {"test_cli": {"option": {}}}
    subcommands = {
        "test_cli": {"subcommand": {"description": "Subcommand description"}}
    }
    OPT = await hub.pop.config.load(
        cli=cli,
        cli_config=cli_config,
        subcommands=subcommands,
        global_clis=[],
        parser_args=["subcommand"],
    )
    assert OPT["pop"]["subparser"] == "subcommand"


async def test_prioritize_with_env_vars(hub):
    # Test prioritizing with environment variables
    cli = "test_cli"
    cli_opts = {}
    config = {"test_cli": {"option": {"os": "TEST_OPTION", "default": "default"}}}
    hub.lib.os.environ["TEST_OPTION"] = "env_value"
    OPT = await hub.pop.config.prioritize(
        cli=cli, cli_opts=cli_opts, config=config, config_file_data={}, global_clis=[]
    )
    assert OPT["test_cli"]["option"] == "env_value"


async def test_load_basic(hub):
    # Test loading with minimal arguments
    OPT = await hub.pop.config.load()
    assert isinstance(OPT, hub.lib.cpop.data.ImmutableNamespaceDict)
    assert "pop" in OPT
    assert "subparser" in OPT["pop"]
    assert "global_clis" in OPT["pop"]
    assert "log" in OPT


async def test_load_with_cli(hub):
    # Test loading with a CLI
    cli = "test_cli"
    cli_config = {"test_cli": {"option": {}}}
    config = {"test_cli": {"option": {"default": "default", "os": "TEST_OPTION"}}}

    hub.lib.os.environ.pop("TEST_OPTION", None)
    # Test the default
    OPT = await hub.pop.config.load(
        cli=cli,
        cli_config=cli_config,
        config=config,
        subcommands={},
        global_clis=[],
        parser_args=[],
    )
    assert OPT[cli]["option"] == "default"

    # Test that os vars overrides default
    hub.lib.os.environ["TEST_OPTION"] = "os"
    OPT = await hub.pop.config.load(
        cli=cli,
        cli_config=cli_config,
        config=config,
        subcommands={},
        global_clis=[],
        parser_args=[],
    )
    assert OPT[cli]["option"] == "os"

    # Test that cli supersedes all
    OPT = await hub.pop.config.load(
        cli=cli,
        cli_config=cli_config,
        config=config,
        subcommands={},
        global_clis=[],
        parser_args=["--option=cli"],
    )
    assert OPT[cli]["option"] == "cli"


async def test_parse_cli(hub):
    # Test parsing CLI options
    cli = "test_cli"
    active_cli = {"option": {"subcommands": ["__global__"]}}
    subcommands = {"subcommand": {}}
    parser_args = ("--option", "cli_value", "subcommand", "--option", "sub_cli_value")
    cli_opts = await hub.pop.config.parse_cli(
        cli=cli, active_cli=active_cli, subcommands=subcommands, parser_args=parser_args
    )
    assert cli_opts["option"] == "sub_cli_value"
    assert cli_opts["SUBPARSER"] == "subcommand"


async def test_prioritize(hub):
    # Test prioritizing configuration options
    cli = "test_cli"
    cli_opts = {"option": "cli_value"}
    config = {"test_cli": {"option": {"default": "config_value"}}}
    config_file_data = {"test_cli": {"option": "file_value"}}
    OPT = await hub.pop.config.prioritize(
        cli=cli,
        cli_opts=cli_opts,
        config=config,
        config_file_data=config_file_data,
        global_clis=["pop", "log"],
    )
    # CLI options have the highest priority
    assert OPT["test_cli"]["option"] == "cli_value"



"""
Contract Context
"""

__virtualname__ = "ctx_update"


async def pre(hub, ctx):
    """ """
    assert ctx.args == [hub, True]
    # Let's replace the context arguments
    ctx.args[1] = False
    assert ctx.args == [hub, False]


async def call_test_call(hub, ctx):
    """ """
    assert ctx.args == [hub, False]
    return "contract executed"


async def post(hub, ctx):
    """ """
    assert ctx.args == [hub, False]
    return ctx.ret



repeats = 10000


async def test_direct(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    for i in range(repeats):
        await hub.mods.test.ping()


async def test_contract(hub):
    await hub.pop.sub.add(pypath=["tests.cmods"])
    for i in range(repeats):
        await hub.cmods.ctest.cping()


async def test_via_fqn(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    for i in range(repeats):
        await hub.mods.test.fqn()



async def test_attr(hub):
    # Set an initial value with setattr
    hub.foo = 1
    # Invoke getattr
    hub.foo
    # Change the value with setattr
    hub.foo = 2
    # The change should be successful
    assert hub.foo == 2



from aiopath import Path

current_dir = Path(globals()["__file__"]).parent


async def test_masked_contracts(hub):
    await hub.pop.sub.add(
        pypath=["tests.regression.contract_masking.sub"],
        contracts_static=[
            current_dir / "contract1",
            current_dir / "contract2",
        ],
    )
    val = await hub.sub.test.func()
    assert len(val) == 2
    assert val == ["contract2", "contract1"]


async def test_masked_recursive_contracts(hub):
    await hub.pop.sub.add(
        pypath=["tests.regression.contract_masking.sub"],
        recursive_contracts_static=[
            current_dir / "contract1",
            current_dir / "contract2",
        ],
    )
    val = await hub.sub.test.func()
    assert len(val) == 2
    assert val == ["contract2", "contract1"]



__virtualname__ = "coro"


try:
    import tornado.gen

    HAS_TORNADO = True
except ImportError:
    HAS_TORNADO = False


async def asyncio_demo(hub):
    return True


if HAS_TORNADO:

    @tornado.gen.coroutine
    def tornado_demo(hub):
        return False



from cpop.verify import sig


async def test_sig_default_values():
    def sig_default(hub, name=None):
        pass

    def no_default(hub, name):
        pass

    result = sig(no_default, sig_default)
    assert 'no_default: Parameter "name" must have a default value' in result[0]
    assert "Enforcing signature: " in result[1]

    def default(hub, name=None):
        pass

    result = sig(default, sig_default)
    assert [] == result


async def test_sig_no_default_values():
    def sig_default(hub, name):
        pass

    def no_default(hub, name):
        pass

    result = sig(no_default, sig_default)
    assert [] == result

    def default(hub, name=None):
        pass

    result = sig(default, sig_default)
    assert 'default: Parameter "name" cannot have a default value' in result[0]
    assert "Enforcing signature: " in result[1]


async def test_sig_no_default_after_star():
    def sig_star(hub, name, *, kwarg_only):
        pass

    def kwarg_not_defined(hub, name, *, kwarg_only):
        pass

    result = sig(kwarg_not_defined, sig_star)
    assert result == []


async def test_sig_error_shows_sig_path():
    def sig_test():
        pass

    def test(param):
        pass

    result = sig(test, sig_test)

    assert "Enforcing signature: " in result[1]
    assert result[1].endswith("test_sigs.py::sig_test")
    assert len(result[1]) > 46



"""
Used to resolve resolutions to paths on the hub
"""


def last(hub, ref) -> object:
    """
    Takes a string that references the desired ref and returns the last object
    called out in that ref
    """
    refs = path(hub, ref)
    return refs.pop()


def path(hub, ref) -> list[object]:
    """
    Retuns a list of references up to the named ref
    """
    ret = [hub]

    if isinstance(ref, str):
        ref = ref.split(".")

    root = hub
    for chunk in ref:
        root = getattr(root, chunk)
        ret.append(root)
    return ret


def find(hub, ref: str) -> object:
    """
    Take a string that represents an attribute nested underneath the hub.
    Parse the string and retrieve the object form the hub.

    Args:
        hub (cpop.hub.Hub): The global namespace.
        ref (str): A string separated by "." with each part being a level deeper into objects on the hub.

    Returns:
        any: The object found on the hub
    """
    # Get the named reference from the hub
    finder = hub
    parts = ref.split(".")
    for p in parts:
        if not p:
            continue
        try:
            # Grab the next attribute in the reference
            finder = getattr(finder, p)
            continue
        except AttributeError:
            try:
                # It might be a dict-like object, try getitem
                finder = finder.__getitem__(p)
                continue
            except TypeError:
                # It might be an iterable, if the next part of the ref is a digit try to access the index
                if p.isdigit() and isinstance(finder, hub.lib.typing.Iterable):
                    finder = tuple(finder).__getitem__(int(p))
                    continue
            raise
    return finder



import os


async def __func_alias__(hub):
    out = {
        "list_": "list",
    }

    out["cpu_count"] = os.cpu_count
    out["λ"] = lambda: 0
    out["already_has_a_hub"] = hub.pop.sub.add
    out["async_func_wrap"] = _async_func

    return out


async def _async_func():
    return True


async def list_(hub):
    return ["list"]



async def pre(hub, ctx):
    print("init-pre")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["init-pre"]


async def post(hub, ctx):
    print("init-post")
    ctx.ret.append("init-post")


async def call(hub, ctx):
    print("init-pre-call")
    result = ctx.extra + ["init-pre-call", await ctx.func(hub=hub), "init-post-call"]
    print("init-post-call")
    return result


async def pre_test_fn(hub, ctx):
    print("init-pre-test-fn")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["init-pre-test-fn"]


async def post_test_fn(hub, ctx):
    print("init-post-test-fn")
    ctx.ret.append("init-post-test-fn")


async def call_test_fn(hub, ctx):
    print("init-pre-call-test-fn")
    result = ctx.extra + [
        "init-pre-call-test-fn",
        await ctx.func(hub=hub),
        "init-post-call-test-fn",
    ]
    print("init-post-call-test-fn")
    return result



async def pre_func1(hub, ctx):
    return False


async def pre_func2(hub, ctx):
    return False, "custom message"


async def pre_afunc1(hub, ctx):
    return False


async def pre_afunc2(hub, ctx):
    return False, "custom message"


async def pre_gen1(hub, ctx):
    return False


async def pre_gen2(hub, ctx):
    return False, "custom message"


async def pre_agen1(hub, ctx):
    return False


async def pre_agen2(hub, ctx):
    return False, "custom message"



"""
Fail to load to test load errors making sure to find the module under
the defined __virtualname__
"""

__virtualname__ = "virtual_bad"


async def __virtual__(hub):
    return "Failed to load virtual bad"


async def func(hub):
    return "wha?"



async def test_contract_context(hub):
    await hub.pop.sub.add(
        pypath=["tests.mods.contract_ctx"],
        subname="mods",
        contracts_pypath=["tests.contracts"],
    )
    assert await hub.mods.ctx.test() == "contract executed"
    # Multiple calls have the same outcome
    assert await hub.mods.ctx.test() == "contract executed"


async def test_contract_context_update_call(hub):
    # if a pre modifies args, make sure they persist when called via 'call' function
    await hub.pop.sub.add(
        pypath=["tests.mods.contract_ctx"],
        subname="mods",
        contracts_pypath=["tests.contracts"],
    )
    assert await hub.mods.ctx_update.test_call(True) == "contract executed"
    # Multiple calls have the same outcome
    assert await hub.mods.ctx_update.test_call(True) == "contract executed"


async def test_contract_context_update_direct(hub):
    # if a pre modifies args, make sure they persist when called directly
    await hub.pop.sub.add(
        pypath=["tests.mods.contract_ctx"],
        subname="mods",
        contracts_pypath=["tests.contracts"],
    )
    assert await hub.mods.ctx_update.test_direct(True) is False
    assert await hub.mods.ctx_update.test_direct(True) is False



"""
Load the files detected from the scanner
"""

import asyncio
import importlib.machinery
import importlib.util
import inspect
import os
import sys
import types
from collections.abc import Callable
from typing import Any
from typing import Union

import cpop.contract
import cpop.data

BASE_TYPES = (int, float, str, bytes, bool, type(None))


cdef class LoadError(Exception):
    cdef public dict edict
    cdef public object traceback

    def __init__(self, msg, exception=None, traceback=None, verror=None):
        self.edict = {
            "msg": msg,
            "exception": exception,
            "verror": verror,
        }
        self.traceback = traceback


def load_mod(modname: str, form: str, path: str) -> "LoadedMod":
    """
    Load a single module
    :param modname: The name of the module to get from the loader
    :param form: The name of the loader module
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    this = sys.modules[__name__]
    return getattr(this, form)(modname, path)


def python(modname: str, path: str) -> "LoadedMod" or LoadError:
    """
    Attempt to load the named python modules
    :param modname: The name of the module to get from the loader
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    if modname in sys.modules:
        return sys.modules[modname]
    spec = importlib.util.spec_from_file_location(modname, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[modname] = module
    module.__file__ = path
    spec.loader.exec_module(module)
    return module


def _base_name(bname: str, mod: "LoadedMod") -> tuple[str, str]:
    """
    Find the basename and alias of a loader module
    :param bname: The base name of the mod's path
    :param mod: A loader module or a LoadError if the module didn't load
    """
    base_name = os.path.basename(bname)
    return base_name, getattr(mod, "__virtualname__", base_name)


async def _load_virtual(
    hub,
    mod: "LoadedMod" or LoadError,
    bname: str,
    vtype: str,
    ref: str,
) -> dict[str, Any]:
    """
    Run the virtual function to name the module and check for all loader
    errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    :param vtype: The name of the virtual function to call on the module I.E. __virtual__ or __sub_virtual__
    """
    base_name, name = _base_name(bname, mod)

    if not hasattr(mod, vtype):
        # No __virtual__ processing is required.
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}

    try:
        virtual_func = cpop.contract.Contracted(
            hub,
            contracts=[],
            func=getattr(mod, vtype),
            ref=f"{ref}.{name}",
            parent=mod,
            name="__virtual__",
        )
        vret = await virtual_func()
    except Exception as e:
        vret = False, str(e)

    verror = vret
    if isinstance(vret, tuple):
        if len(vret) > 1:
            verror = vret[1]
        vret = vret[0]

    if vret:
        # No problems occurred, module is allowed to load
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}
    else:
        # __virtual__ explicitly disabled the loading of this module
        err = LoadError(f"Module {bname} returned virtual FALSE", verror=verror)
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {"name": base_name, "vname": name, "error": err}


async def load_virtual(hub, mod: "LoadedMod" or LoadError, bname: str, ref: str) -> dict[str, Any]:
    """
    Run the __virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    return await _load_virtual(hub, mod, bname, "__virtual__", ref)


async def load_sub_virtual(
    hub, mod: "LoadedMod" or LoadError, bname: str, ref: str
) -> dict[str, Any]:
    """
    Run the __sub_virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    _, name = _base_name(bname, mod)
    if name != "init":
        return {"name": name}
    return await _load_virtual(hub, mod, bname, "__sub_virtual__", ref)


async def mod_init(sub, mod: "LoadedMod", mod_name: str):
    """
    Process module's __init__ function if defined
    :param sub: The pop object that contains the loaded module data
    :param mod: A loader modul
    :param mod_name: The name of the module to get from the loader
    """
    if "__init__" in dir(mod):
        init = cpop.contract.Contracted(
            sub._hub,
            contracts=[],
            func=mod.__init__,
            ref=f"{sub._ref}.{mod_name}",
            parent=mod,
            name="__init__",
        )
        await init()


def sub_alias(this_sub, mod: "LoadedMod", mod_name: str):
    """
    Check the sub alias settings and apply the alias names locally so they can be gathered into the
        higher level object on the hub
    :param this_sub: The pop object that contains the loaded module data
    :param mod: A loader module
    :param mod_name: The name of the module to get from the loader
    """
    if mod_name == "init":
        alias = getattr(mod, "__sub_alias__", None)
        if alias:
            this_sub._alias = alias


async def prep_loaded_mod(
    this_sub,
    mod: "LoadedMod",
    mod_name: str,
    contracts: "list[cpop.contract.Wrapper]",
    recursive_contracts: "list[cpop.contract.Wrapper]",
) -> "LoadedMod":
    ordered_contracts = contracts + recursive_contracts

    lmod = this_sub._loaded.get(mod_name, LoadedMod(mod_name, parent=this_sub))

    # getattr(hub, ref) should resolve to this module
    ref = f"{this_sub._ref}.{mod_name}"

    sub_alias(this_sub, mod, mod_name)
    func_alias_dict: dict[str, Union[str, Callable, cpop.contract.Contracted]] = {}

    __func_alias__: Union[Callable, dict[str, Union[str, Callable]]] = getattr(mod, "__func_alias__", {})
    if asyncio.iscoroutinefunction(__func_alias__):
        func_alias_dict.update(await __func_alias__(this_sub._hub))
    else:
        func_alias_dict.update(__func_alias__)

    for attr in getattr(mod, "__load__", dir(mod)):
        func_alias: Union[str, Callable] = func_alias_dict.get(attr, attr)
        func = getattr(mod, attr)
        name = func_alias if isinstance(func_alias, str) else attr

        if not this_sub._omit_vars and not (
                    inspect.isfunction(func) or
                    inspect.isclass(func) or
                    type(func).__name__ == "cython_function_or_method"
                ):
            setattr(lmod, name, func)
            continue

        if attr.startswith(tuple(this_sub._omit_start)) or attr.endswith(tuple(this_sub._omit_end)):
            continue

        if inspect.isfunction(func) or inspect.isbuiltin(func) or type(func).__name__ == "cython_function_or_method":
            obj = cpop.contract.create_contracted(this_sub._hub, ordered_contracts, func, ref, name, parent=lmod)
            if not this_sub._omit_func and (not this_sub._pypath or func.__module__.startswith(mod.__name__)):
                lmod._funcs[name] = obj
        elif not this_sub._omit_class and inspect.isclass(func) and func.__module__.startswith(mod.__name__):
            lmod._classes[name] = func

    for attr, func_alias in func_alias_dict.items():
        if isinstance(func_alias, cpop.contract.Contracted):
            obj = func_alias
        elif isinstance(func_alias, Callable):
            obj = cpop.contract.create_contracted(
                this_sub._hub,
                ordered_contracts,
                func_alias,
                ref,
                attr,
                parent=lmod,
                implicit_hub=False
            )
        else:
            continue
        lmod._funcs[attr] = obj

    return lmod


class LoadedMod(types.ModuleType):
    """
    The LoadedMod class allows for the module loaded onto the sub to return
    custom sequencing, for instance it can be iterated over to return all
    functions
    """

    def __init__(self, name: str, parent: cpop.hub.Sub):
        super().__init__(name)
        cdef dict vars = {}
        cdef dict funcs = {}
        cdef dict classes = {}
        self._attrs = cpop.data.MultidictCache([funcs, classes, vars], parent=parent)
        self._vars = vars
        self._funcs = funcs
        self._classes = classes

    def __getattr__(self, item: str):
        if item == "__":
            return self._attrs.__
        elif item.startswith("_"):
            return self.__getattribute__(item)
        try:
            return self._attrs[item]
        except KeyError as e:
            raise AttributeError(e)

    def __getitem__(self, item: str):
        return self._attrs[item]

    def __setattr__(self, item: str, value):
        if item.startswith("_"):
            object.__setattr__(self, item, value)
            return
        elif isinstance(value, types.FunctionType):
            self._funcs[item] = value
        elif isinstance(value, type):
            self._classes[item] = value
        else:
            self._vars[item] = value
        self._attrs._clear()

    def __getstate__(self):
        return {
            "vars": {k: v for k, v in self._vars.items() if isinstance(v, BASE_TYPES)}
        }

    def __setstate__(self, state: dict[str, any]):
        self._vars.update(state["vars"])

    def __iter__(self):
        return iter(self._attrs)

    @property
    def _ref(self) -> str:
        """
        Return a full path on the hub to this mod
        """
        ref = self.__name__
        root = self._attrs._parent
        while hasattr(root, "_subname"):
            ref = f"{root._subname}.{ref}"
            root = root._parent
        return ref



async def func1(hub):
    raise Exception("Pre Contract did not fail")


async def func2(hub):
    raise Exception("Pre Contract did not fail")


async def afunc1(hub):
    raise Exception("Pre Contract did not fail")


async def afunc2(hub):
    raise Exception("Pre Contract did not fail")


async def gen1(hub):
    if True:
        raise Exception("Pre Contract did not fail")
    else:
        yield


async def gen2(hub):
    if True:
        raise Exception("Pre Contract did not fail")
    else:
        yield


async def agen1(hub):
    if True:
        raise Exception("Pre Contract did not fail")
    else:
        yield


async def agen2(hub):
    if True:
        raise Exception("Pre Contract did not fail")
    else:
        yield



"""
Example contract
"""

__virtualname__ = "test"


async def functions(hub):
    """
    Return the functions
    """
    return ("ping", "demo")


async def pre_ping(hub, ctx):
    """ """
    if len(ctx.args) != 1:
        raise Exception("ping does not take args!")
    if ctx.kwargs:
        raise Exception("ping does not take kwargs!")


async def call_ping(hub, ctx):
    """ """
    print("calling!")
    return await ctx.func(*ctx.args, **ctx.kwargs)


async def post_ping(hub, ctx):
    """ """
    print("Calling Post!")
    if not isinstance(ctx.ret, dict):
        raise Exception("MUST BE DICT!!")



async def pre(hub, ctx):
    print("rc_sub1-recursive-pre")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["rc_sub1-recursive-pre"]


async def post(hub, ctx):
    print("rc_sub1-recursive-post")
    ctx.ret.append("rc_sub1-recursive-post")


async def call(hub, ctx):
    print("rc_sub1-recursive-pre-call")
    result = ctx.extra + [
        "rc_sub1-recursive-pre-call",
        await ctx.func(hub=hub),
        "rc_sub1-recursive-post-call",
    ]
    print("rc_sub1-recursive-post-call")
    return result



"""
Find directories
"""

import importlib.resources
import os
import aiopath
import sys
from collections import defaultdict
from collections.abc import Iterable

import cpop.data
import msgpack
import yaml


async def dir_list(pypath: list[str] = None, static: list[str] = None) -> list[aiopath.Path]:
    """
    Return the directories to look for modules in, pypath specifies files
    relative to an installed python package, static is for static dirs
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    """
    ret = set()
    if pypath:
        for path in pypath:
            try:
                mod = importlib.import_module(path)
            except ModuleNotFoundError:
                ...
            for m_path in mod.__path__:
                # If we are inside of an executable the path will be different
                ret.add(aiopath.Path(m_path))
    if static:
        ret.update(aiopath.Path(dir_) for dir_ in static)
    return sorted(ret)


async def inline_dirs(dirs: Iterable[str], subdir: str) -> list[aiopath.Path]:
    """
    Look for the named subdir in the list of dirs
    :param dirs: The names of configured dynamic dirs
    :param subdir: The name of the subdir to check for in the list of dynamic dirs
    :return An extended list of dirs that includes the found subdirs
    """
    ret = set()
    for dir_ in dirs:
        check = aiopath.Path(dir_) / subdir
        if await check.is_dir():
            ret.add(check)
    return sorted(ret)


async def dynamic_dirs():
    """
    Iterate over the available python package imports and look for configured
    dynamic dirs in pyproject.toml
    """
    dirs = set()
    for dir_ in sys.path:
        if not dir_:
            continue
        path = aiopath.Path(dir_)
        if not await path.is_dir():
            continue
        async for sub in path.iterdir():
            full = path / sub
            if str(sub).endswith(".egg-link"):
                async with full.open() as rfh:
                    dirs.add(aiopath.Path((await rfh.read()).strip()))
            elif await full.is_dir():
                dirs.add(full)

    # Set up the _dynamic return
    ret = cpop.data.NamespaceDict(
        dyne=cpop.data.NamespaceDict(),
        config=cpop.data.NamespaceDict(),
        imports__=cpop.data.NamespaceDict(),
    )

    # Iterate over namespaces in sys.path
    for dir_ in dirs:
        # Prefer msgpack configuration if available
        config_msgpack = dir_ / "config.msgpack"
        config_yaml = dir_ / "config.yaml"

        config_file = config_msgpack if await config_msgpack.is_file() else None
        if not config_file and await config_yaml.is_file():
            config_file = config_yaml

        if not config_file:
            # No configuration found, continue with the next directory
            continue

        dynes, configs, imports = await parse_config(config_file)
        if dynes:
            cpop.data.update(ret.dyne, dynes, merge_lists=True)
        if configs:
            cpop.data.update(ret.config, configs, merge_lists=True)
        if imports:
            cpop.data.update(ret.imports__, imports)

    return ret


async def parse_config(config_file: aiopath.Path) -> tuple[dict[str, object], dict[str, object], set[str]]:
    dyne = defaultdict(lambda: cpop.data.NamespaceDict(paths=set()))
    config = cpop.data.NamespaceDict(
        config=cpop.data.NamespaceDict(),
        cli_config=cpop.data.NamespaceDict(),
        subcommands=cpop.data.NamespaceDict(),
    )
    imports = cpop.data.NamespaceDict()

    if not await config_file.is_file():
        return dyne, config, imports

    async with config_file.open("rb") as f:
        file_contents = await f.read()
        if config_file.suffix == ".yaml" or config_file.suffix == ".yml":
            pop_config = yaml.safe_load(file_contents) or {}
        elif config_file.suffix == ".msgpack":
            pop_config = msgpack.unpackb(file_contents) or {}
        else:
            raise ValueError("Unsupported file format")

    # Gather dynamic namespace paths for this import
    for name, paths in pop_config.get("dyne", {}).items():
        for path in paths:
            ref = config_file.parent / path.replace(".", os.sep)
            dyne[name]["paths"].add(ref)

    # Get config sections
    for section in ["config", "cli_config", "subcommands"]:
        section_data = pop_config.get(section)
        if not isinstance(section_data, dict):
            continue
        for namespace, data in section_data.items():
            if data is None:
                continue
            config[section].setdefault(namespace, cpop.data.NamespaceDict()).update(data)

    # Handle python imports
    for imp in pop_config.get("import", []):
        base = imp.split(".", 1)[0]
        if base not in imports:
            try:
                imports[base] = importlib.import_module(base)
            except ModuleNotFoundError:
                ...
        if "." in imp:
            try:
                importlib.import_module(imp)
            except ModuleNotFoundError:
                ...

    for name in dyne:
        dyne[name]["paths"] = sorted(dyne[name]["paths"])

    return dyne, config, imports



import os

from cpop.scanner import scan

__virtualname__ = "test"
__contracts__ = "test"
__func_alias__ = {"ping_": "ping"}


async def ping_(hub):
    return {}


async def demo(hub):
    return False


async def this(hub):
    return await hub._.ping()


async def fqn(hub):
    return await hub.mods.test.ping()


async def module_level_non_aliased_ping_call(hub):
    return await ping_()  # pylint: disable=no-value-for-parameter


async def module_level_non_aliased_ping_call_fw_hub(hub):
    return await ping_(hub)


async def attr(hub):
    return True


attr.bar = True
attr.func = True


async def call_scan(hub):
    # If scan has been loaded as a Contract, the call below will throw a TypeError because
    # we'll also pass hub
    scan([os.path.dirname(__file__)])
    return True


async def double_underscore(hub):
    assert hub.__ is hub.mods
    assert hub.___ is hub
    assert hub.______ is hub



import inspect

import cpop.data
from cpop.contract import Contracted


async def test_contracted_shortcut(hub):
    async def f(hub):
        pass

    c = Contracted(hub=hub, contracts=[], func=f, ref="", name="", parent=None)
    c.contract_functions["pre"] = [
        None
    ]  # add some garbage so we raise if we try to evaluate contracts

    await c()


async def test_contracted_inspect():
    def f(hub, p1, p2=None):
        pass

    c = Contracted(None, [], f, "", "", None)

    assert str(inspect.signature(c)) == str(inspect.signature(f))
    assert str(inspect.signature(c)) == "(hub, p1, p2=None)"



async def _hub(hub):
    await hub.pop.sub.add(pypath=["tests.integration.contracted.mods"])
    return hub


async def test_two_hubs(hub):
    h = await _hub(hub)

    # we should be able to call a function with two hubs as parameters
    await h.mods.contracted_access.two_hubs(h)



async def __virtual__(hub):
    return getattr(hub, "LOAD_PASS", False), "Plugin not enabled"


async def args(hub, no_type_param, typed_param: str, *_): ...


async def kwargs(hub, no_type_param=None, typed_param: str = "", **__): ...


async def args_kwargs(
    hub,
    no_type_param,
    typed_param: str,
    no_type_param_default=None,
    typed_param_default: str = "",
    *_,
    **__,
): ...


async def async_func(hub): ...



async def pre(hub, ctx):
    print("regular-pre")
    ctx.extra = (getattr(ctx, "extra", None) or []) + ["regular-pre"]


async def post(hub, ctx):
    print("regular-post")
    ctx.ret.append("regular-post")


async def call(hub, ctx):
    print("regular-pre-call")
    result = ctx.extra + ["regular-pre-call", ctx.func(hub=hub), "regular-post-call"]
    print("regular-post-call")
    return result



import asyncio
import inspect
import sys
import traceback
from collections.abc import Iterable
from cpython cimport dict

cdef class ContractedContext:
    """
    Contracted function calling context
    """
    cdef:
        public object func
        public list args
        public dict[str, object] kwargs
        public str ref
        public str __name__
        public object ret
        object signature
        dict[str, object] __cache

    def __init__(
        self,
        func: object,
        args: tuple,
        kwargs: dict,
        signature: object,
        ref: str,
        name: str,
        ret=None,
        cache=None,
    ):
        if cache is None:
            cache = {}

        self.func = func
        self.args = list(args)
        self.kwargs = kwargs
        self.signature = signature
        self.ref = ref
        self.__name__ = name
        self.ret = ret
        self.__cache = cache

    @property
    def __signature__(self):
        return self.signature

    @property
    def cache(self):
        return self.__cache

    def get_arguments(self):
        """
        Return a dictionary of all arguments that will be passed to the function and their
        values, including default arguments.
        """
        if "__bound_signature__" not in self.__cache:
            self.__cache["__bound_signature__"] = self.signature.bind(
                *self.args, **self.kwargs
            )
            # Apply any default values from the signature
            self.__cache["__bound_signature__"].apply_defaults()
        return self.__cache["__bound_signature__"].arguments


def load_contract(
    contracts,
    default_contracts: Iterable[str],
    mod: object,
    name: str,
) -> list[object]:
    """
    return a Contract object loaded up
    Dynamically create the correct Contracted type
    :param contracts: Contracts functions to add to the sub
    :param default_contracts: The contracts that have been marked as defaults
    :param mod: A loader module
    :param name: The name of the module to get from the loader
    """
    raws = []
    loaded_contracts = []
    if hasattr(contracts, name):
        loaded_contracts.append(name)
        raws.append(getattr(contracts, name))
    if hasattr(contracts, "init"):
        if "init" not in loaded_contracts:
            loaded_contracts.append("init")
            raws.append(getattr(contracts, "init"))
    if hasattr(mod, "__contracts__"):
        cnames = getattr(mod, "__contracts__")
        for cname in cnames:
            if cname in contracts:
                loaded_contracts.append(cname)
                raws.append(getattr(contracts, cname))
    return raws


cdef class Contracted:
    """
    This class wraps functions that have a contract associated with them
    and executes the contract routines
    """
    cdef:
        public bint _has_contracts
        cdef public bint implicit_hub
        public dict[str, list[object]] contract_functions
        public object hub
        public object func
        public object signature
        cdef public str __name__
        cdef public str ref
        list _sig_errors
        list contracts
        object __wrapped__
        object parent

    def __init__(
                self,
                hub,
                contracts: list,
                func: object,
                ref: str,
                name: str,
                parent: object,
                implicit_hub: bool = True
            ):
        self.__name__ = name
        self.__wrapped__ = func
        self._sig_errors = []
        self.contracts = contracts or []
        self.parent = parent
        self.func = func
        self.hub = hub
        self.implicit_hub = implicit_hub
        self.ref = f"{ref}.{name}"
        # Manually define the function signature if needed
        self.signature = inspect.signature(self.func)

        self._load_contracts()

    @property
    def __doc__(self):
        return self.func.__doc__

    @property
    def __(self):
        return self.parent

    @property
    def __signature__(self):
        return self.signature

    @property
    def __dict__(self):
        return self.func.__dict__

    cdef _get_contracts_by_type(self, contract_type: str):
        matches = []
        fn_contract_name = f"{contract_type}_{self.__name__}"
        for contract in self.contracts:
            if hasattr(contract, fn_contract_name):
                matches.append(getattr(contract, fn_contract_name))
            if hasattr(contract, contract_type):
                matches.append(getattr(contract, contract_type))

        if contract_type == "post":
            matches.reverse()

        return matches

    cdef _load_contracts(self):
        # if Contracted - only allow regular pre/post
        # if ContractedAsync - allow coroutines and functions

        self.contract_functions = {
            "pre": self._get_contracts_by_type("pre"),
            "call": self._get_contracts_by_type("call")[:1],
            "post": self._get_contracts_by_type("post"),
        }
        self._has_contracts = sum(len(funcs) for funcs in self.contract_functions.values()) > 0

    async def __call__(self, *args, **kwargs):
        async with CallStack(self):
            if self.implicit_hub:
                args = (self.hub,) + args

            if not self._has_contracts:
                ret = self.func(*args, **kwargs)
                if asyncio.iscoroutine(ret):
                    ret = await ret
                return ret
            contract_context = ContractedContext(
                self.func, args, kwargs, self.signature, self.ref, self.__name__
            )

            # Process pre contracts
            for fn in self.contract_functions["pre"]:
                pre_ret = await fn(contract_context)

                await self.hub.pop.contract.process_pre_result(pre_ret, fn, self)

            # Call the one call contract
            if self.contract_functions["call"]:
                ret = self.contract_functions["call"][0](contract_context)
            else:
                ret = self.func(*contract_context.args, **contract_context.kwargs)

            if asyncio.iscoroutine(ret):
                ret = await ret

            # Handle post contracts
            for fn in self.contract_functions["post"]:
                contract_context.ret = ret
                post_ret = await fn(contract_context)
                if post_ret is not None:
                    ret = post_ret

            return ret

    def __getstate__(self):
        return {
            "ref": self.ref,
            "name": self.__name__,
            "implicit_hub": self.implicit_hub,
            "contracts": self.contracts,
        }

    def __setstate__(self, state: dict[str, object]):
        self.ref = state["ref"]
        self.__name__ = state["name"]
        self.func = self.hub[self.ref][self.__name__].func
        self.implicit_hub = state["implicit_hub"]
        self.contracts = state["contracts"]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.ref}>"


class ContractedAsyncGen(Contracted):
    async def __call__(self, *args, **kwargs):
        async with CallStack(self):
            if self.implicit_hub:
                args = (self.hub,) + args

            if not self._has_contracts:
                async for chunk in self.func(*args, **kwargs):
                    yield chunk
                return
            contract_context = ContractedContext(
                self.func, args, kwargs, self.signature, self.ref, self.__name__
            )

            for fn in self.contract_functions["pre"]:
                pre_ret = await fn(contract_context)

                await self.hub.pop.contract.process_pre_result(pre_ret, fn, self)
            chunk = None
            if self.contract_functions["call"]:
                async for chunk in self.contract_functions["call"][0](contract_context):
                    yield chunk
            else:
                async for chunk in self.func(
                    *contract_context.args, **contract_context.kwargs
                ):
                    yield chunk
            ret = chunk
            for fn in self.contract_functions["post"]:
                contract_context.ret = ret
                post_ret = await fn(contract_context)
                if post_ret is not None:
                    ret = post_ret


class CallStack:
    """
    A wrapper for functions to add and remove their context from the stack securely
    """
    def __init__(self, contract: Contracted):
        self.contract = contract

    async def __aenter__(self):
        # Get the caller's frame
        frame = sys._getframe(1)
        stack_summary = traceback.StackSummary.extract(traceback.walk_stack(frame))
        # Push the ref and the stack summary onto the call stack
        self.contract.hub._call_stack_push(self.contract.ref, stack_summary)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Remove the entry from the call stack
        self.contract.hub._call_stack_pop()


def create_contracted(
    hub,
    contracts: list["cpop.loader.LoadedMod"],
    func,
    ref: str,
    name: str,
    parent: object,
    implicit_hub: bool = True,
) -> Contracted:
    """
    Dynamically create the correct Contracted type
    :param hub: The redistributed pop central hub
    :param contracts: Contracts functions to add to the sub
    :param func: The contracted function to call
    :param ref: The reference to the function on the hub
    :param name: The name of the module to get from the loader
    :param parent: The object on the namespace above this contract
    :param implicit_hub: True if a hub should be implicitly injected into the "call" method
    """
    if inspect.isasyncgenfunction(func):
        return ContractedAsyncGen(hub, contracts, func, ref, name, implicit_hub)
    else:
        return Contracted(hub, contracts, func, ref, name, parent, implicit_hub)



async def two_hubs(hub1, hub2):
    return


async def hub2_dereferenced_call(hub1, hub2):
    hub1.mods.contracted_access.hub1_called = True
    hub2._.hub2_call()


async def hub2_direct_call(hub1, hub2):
    hub1.mods.contracted_access.hub1_called = True
    hub2_call(hub2)


async def hub2_direct_call_kwargs(hub1, hub2):
    hub1.mods.contracted_access.hub1_called = True
    hub2_call(h=hub2)


async def hub2_call(h):
    h.mods.contracted_access.hub2_called = True



from . import loader
from . import hub
from . import data
from . import contract
from .exc import *

LoadedMod = loader.LoadedMod
Hub = hub.Hub
MultidictCache = data.MultidictCache
NamespaceDict = data.NamespaceDict
Sub = hub.Sub
AsyncSub = hub.AsyncSub
Contracted = contract.Contracted
ContractedContext = contract.ContractedContext
ImmutableNamespaceDict = data.ImmutableNamespaceDict


__all__ = [
    "AsyncSub",
    "Contracted",
    "Hub",
    "ImmutableNamespaceDict",
    "LoadedMod",
    "MultidictCache",
    "NamespaceDict",
    "Sub",
]



import random


async def callback(hub, payload):
    if "payload" in payload:
        hub.set_me = payload["payload"]["ret"]
    return "foo"


async def ret(hub):
    await hub.proc.worker.ret({"ret": "Returned"})
    return "inline"


async def gen(hub, start, end):
    for x in range(start, end):
        yield x


async def simple_gen(hub, start, end):
    for i in range(start, end):
        yield i


async def init_lasts(hub):
    hub.LASTS = {}
    return True


async def echo_last(hub):
    """
    Return 2 numbers, the second number is new, the first number is the same that
    was the second number the last time it was called
    """
    next_ = random.randint(0, 50000)
    if "last" not in hub.LASTS:
        hub.LASTS["last"] = 0
    last = hub.LASTS["last"]
    hub.LASTS["last"] = next_
    return last, next_


async def gen_last(hub, num=5):
    for _ in range(num):
        next_ = random.randint(0, 50000)
        if "last" not in hub.LASTS:
            hub.LASTS["last"] = 0
        last = hub.LASTS["last"]
        hub.LASTS["last"] = next_
        yield last, next_



import os


async def _async_func():
    return True


__func_alias__ = {
    "list_": "list",
    "cpu_count": os.cpu_count,
    "λ": lambda: 0,
    "async_func_wrap": _async_func,
}


async def list_(hub):
    return ["list"]



import tempfile

import pytest


@pytest.fixture()
def opts():
    with tempfile.NamedTemporaryFile(suffix=".log") as f:
        yield dict(
            log_datefmt=r"%H:%M:%S",
            log_fmt_logfile=r"%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s",
            log_file=f.name,
            log_plugin="basic",
        )


async def test_trace(hub, opts):
    message = "message"
    opts["log_level"] = "trace"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.trace(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents


async def test_debug(hub, opts):
    message = "message"
    opts["log_level"] = "debug"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.debug(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents


async def test_info(hub, opts):
    message = "message"
    opts["log_level"] = "info"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.info(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents


async def test_error(hub, opts):
    message = "message"
    opts["log_level"] = "error"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.error(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents


async def test_warning(hub, opts):
    message = "message"
    opts["log_level"] = "warning"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.warning(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents


async def test_critical(hub, opts):
    message = "message"
    opts["log_level"] = "critical"
    await hub.pop.sub.add(dyne_name="log")
    await hub.log.init.setup(**opts)
    await hub.log.critical(message)
    with open(opts["log_file"]) as fh:
        contents = fh.read()
        assert message in contents



import cpop.exc
import pytest


@pytest.fixture(scope="function")
async def hub(hub):
    await hub.pop.sub.add(dyne_name="cn")
    yield hub


async def test_pre_fail_sync(hub):
    with pytest.raises(
        cpop.exc.PreContractFailed,
        match=r"Pre contract '\w+' failed for function '\w+'",
    ):
        await hub.cn.test.func1()


async def test_pre_fail_async(hub):
    with pytest.raises(
        cpop.exc.PreContractFailed,
        match=r"Pre contract '\w+' failed for function '\w+'",
    ):
        await hub.cn.test.afunc1()


async def test_pre_fail_gen(hub):
    with pytest.raises(
        cpop.exc.PreContractFailed,
        match=r"Pre contract '\w+' failed for function '\w+'",
    ):
        async for _ in hub.cn.test.gen1():
            ...



import cpop.ref
import pytest


@pytest.fixture
async def hub(hub):
    await hub.pop.sub.add("test")
    await hub.pop.sub.add("subtest", sub=hub.test)
    hub.test.subtest.attr = "value"
    hub.test.subtest.dict_attr = {"key": "dict_value"}
    hub.test.subtest.list_attr = ["list_value"]
    yield hub


async def test_last_str(hub):
    assert cpop.ref.last(hub, "test.subtest.attr") == "value"


def test_path(hub):
    path = cpop.ref.path(hub, "test.subtest.attr")
    assert path == [hub, hub.test, hub.test.subtest, "value"]


def test_find(hub):
    assert cpop.ref.find(hub, "test.subtest.attr") == "value"
    assert cpop.ref.find(hub, "test.subtest.dict_attr.key") == "dict_value"
    assert cpop.ref.find(hub, "test.subtest.list_attr.0") == "list_value"
    with pytest.raises(KeyError):
        cpop.ref.find(hub, "test.subtest.nonexistent")



"""
Used to scan the given directories for loadable files
"""

import os
from typing import Iterable

PY_END = (".py", ".pyc", ".pyo")


def scan(dirs: Iterable[str]) -> dict:
    ret = {}
    ret["python"] = {}
    ret["cython"] = {}
    ret["imp"] = {}
    for dir_ in dirs:
        try:
            for fn_ in os.listdir(dir_):
                _apply_scan(ret, str(dir_), str(fn_))
        except FileNotFoundError:
            ...
    return ret


cdef _apply_scan(dict ret, str dir_, str fn_):
    if fn_.startswith("_"):
        return
    full = os.path.join(dir_, fn_)
    if "." not in full:
        return
    bname = full[: full.rindex(".")]
    if fn_.endswith(PY_END):
        if bname not in ret["python"]:
            ret["python"][bname] = {"path": full}



async def setup(
    hub,
    log_fmt_console: str = None,
    log_datefmt: str = None,
    log_fmt_logfile: str = None,
    log_file: str = None,
    **kwargs,
):
    if log_fmt_console and log_datefmt:
        console_formatter = hub.lib.aiologger.formatters.base.Formatter(
            fmt=log_fmt_console, datefmt=log_datefmt
        )
        console_handler = hub.lib.aiologger.handlers.streams.AsyncStreamHandler(
            formatter=console_formatter, stream=hub.lib.sys.stderr
        )
        hub.log.HANDLER.append(console_handler)

    if log_fmt_logfile and log_file and log_datefmt:
        file_formatter = hub.lib.aiologger.formatters.base.Formatter(
            fmt=log_fmt_logfile, datefmt=log_datefmt
        )
        file_handler = hub.lib.aiologger.handlers.files.AsyncFileHandler(
            filename=log_file
        )
        file_handler.formatter = file_formatter

        hub.log.HANDLER.append(file_handler)



async def setup(hub, **kwargs):
    """
    Given the configuration data set up the logger
    """
    cf = hub.lib.aiologger.formatters.base.Formatter(
        fmt=kwargs["log_fmt_console"], datefmt=kwargs["log_datefmt"]
    )
    ch = hub.lib.aiologger.handlers.streams.AsyncStreamHandler(
        formatter=cf, loop=hub.pop.Loop, stream=hub.lib.sys.stderr
    )
    hub.log.HANDLER.append(ch)

    ff = hub.lib.aiologger.formatters.base.Formatter(
        fmt=kwargs["log_fmt_logfile"], datefmt=kwargs["log_datefmt"]
    )
    fh = hub.lib.aiologger.handlers.files.AsyncTimedRotatingFileHandler(
        kwargs["log_file"]
    )
    fh.formatter = ff
    hub.log.HANDLER.append(fh)



import multiprocessing


async def test_serialize_hub(hub):
    hub.lib.pickle.dumps(hub)


async def test_serialize_sub(hub):
    hub.lib.pickle.dumps(hub.pop)


async def test_serialize_lib(hub):
    hub.lib.pickle.dumps(hub.lib)


def child_process(hub):
    # Try to access or modify the hub object in the child process
    hub.VAR.value += 1
    return hub.VAR


async def test_hub_multiprocessing(hub):
    hub.VAR = multiprocessing.Value("i", 1)
    # Create a multiprocessing Process
    process = multiprocessing.Process(target=child_process, args=(hub,))
    process.start()
    process.join()

    # Check the result
    assert (
        hub.VAR.value == 2
    ), "Hub object should have been incremented in the child process"



"""
Test relative paths like _ and __
"""


async def test_hub_parent(hub):
    assert hub.__ is hub


async def test_sub_parent_hub(hub):
    assert hub.pop.__ is hub


async def test_sub_parent_sub(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.pop.sub.add(pypath=["tests.mods.nest"], sub=hub.mods)
    assert hub.mods.nest.__ is hub.mods


async def test_mod_parent(hub):
    assert hub.pop.sub.__ is hub.pop


async def test_contracted_parent(hub):
    assert hub.pop.sub.add.__ is hub.pop.sub


async def test_current_mod(hub):
    await hub.pop.sub.add(pypath=["tests.mods"])
    await hub.mods.dunder.func()



"""
Contract Context
"""

__virtualname__ = "ctx"


async def pre(hub, ctx):
    """ """
    assert "injected_data" not in ctx.cache
    ctx.cache["injected_data"] = 1


async def call(hub, ctx):
    """ """
    if "injected_data" not in ctx.cache:
        raise AssertionError(
            "'injected_data' was not found in contract call 'ctx.cache'"
        )
    ctx.cache["injected_data"] += 1
    return "contract executed"


async def post_ping(hub, ctx):
    """ """
    if "injected_data" not in ctx.cache:
        raise AssertionError(
            "'injected_data' was not found in contract post 'ctx.cache'"
        )
    assert ctx.cache["injected_data"] == 2
    return ctx.ret
